==> apigw-1.0.3.tgz: package/src/custom-domain.ts <==
import { pascalCaseProperty } from '@tencent-serverless-framework/common';
import { Apigw } from '.';
import { formatProtocolString } from './formatter';
import {
  ApigwBindCustomDomainInputs,
  ApigwBindCustomDomainOutputs,
  ApigwCustomDomain,
} from './interface';

/** 格式化的 API 网关自定义域名信息 */
interface FormattedApigwCustomDomain {
  /** 域名 */
  domain: string;
  /** 协议 */
  protocols: string;
  /** 证书 id */
  certificateId?: string;
  /** 是否是默认映射 */
  isDefaultMapping: boolean;
  /** 映射字典 */
  pathMappingSetDict: Record<string, string>;
  /** 网络类型 */
  netType: string;
}

/**
 * 判断是否是值类型
 *
 * @param obj 要判断的对象
 * @returns 是否是值类型
 */
export function isPrimitive(obj: any): boolean {
  return obj !== Object(obj);
}

/**
 * 深度比较对象是否相等
 *
 * @param obj1 对象 1
 * @param obj2 对象 2
 * @returns 是否相等
 */
export function deepEqual(obj1: any, obj2: any): boolean {
  if (obj1 === obj2) {
    return true;
  }

  if (isPrimitive(obj1) || isPrimitive(obj2)) {
    return obj1 === obj2;
  }

  if (Object.keys(obj1).length !== Object.keys(obj2).length) {
    return false;
  }

  // compare objects with same number of keys
  for (const key in obj1) {
    if (!(key in obj2)) {
      return false;
    }
    if (!deepEqual(obj1[key], obj2[key])) {
      return false;
    }
  }

  return true;
}

/**
 * 获取自定义域名的字典
 *
 * @param domains 自定义域名数组
 * @returns 自定义域名字典
 */
function getCustomDomainFormattedDict(domains: ApigwCustomDomain[]) {
  const domainDict: Record<string, FormattedApigwCustomDomain> = {};
  domains.forEach((d) => {
    const pmDict: Record<string, string> = {};
    for (const pm of d.pathMappingSet ?? []) {
      pmDict[pm.path] = pm.environment;
    }
    domainDict[d.domain] = {
      domain: d.domain,
      certificateId: d.certificateId ?? '',
      protocols: formatProtocolString(d.protocols ?? ''),
      isDefaultMapping: d.isDefaultMapping === false ? false : true,
      pathMappingSetDict: pmDict,
      netType: d.netType ?? 'OUTER',
    };
  });

  return domainDict;
}

/** API 网关组件自定义域名子模块 */
export class CustomDomain {
  /** API 网关组件 */
  apigw: Apigw;
  /**
   * 构造函数
   *
   * @param apigw API 网关组件（依赖注入）
   */
  constructor(apigw: Apigw) {
    this.apigw = apigw;
  }

  /**
   * 获取当前 API 自定义域名字典
   *
   * @param serviceId API ID
   * @returns 自定义域名字典
   */
  async getCurrentCustomDomainsDict(serviceId: string) {
    const res = await this.apigw.client.DescribeServiceSubDomains({
      ServiceId: serviceId,
    });

    const DomainSet = res.Result?.DomainSet;

    const domainDict: Record<string, FormattedApigwCustomDomain> = {};

    for (const d of DomainSet ?? []) {
      const domain: FormattedApigwCustomDomain = {
        domain: d.DomainName,
        protocols: d.Protocol,
        certificateId: d.CertificateId,
        isDefaultMapping: d.IsDefaultMapping,
        netType: d.NetType,
        pathMappingSetDict: {},
      };

      const res_ = await this.apigw.client.DescribeServiceSubDomainMappings({
        ServiceId: serviceId,
        SubDomain: d.DomainName,
      });

      const PathMappingSet = res_.Result?.PathMappingSet;

      PathMappingSet?.map((v) => {
        domain.pathMappingSetDict[v.Path] = v.Environment;
      });

      domainDict[domain.domain] = domain;
    }

    return domainDict;
  }

  /**
   * 解绑 API 网关所有由组件创建的自定义域名（通过查询 oldState）
   *
   * @param serviceId API 网关 ID
   * @param oldCustomDomains 状态中的自定义
   * @param currentDict 当前自定义域名字典
   * @param newDict 要绑定的自定义域名字典
   */
  async unbindCustomDomain(
    serviceId: string,
    oldCustomDomains: ApigwCustomDomain[],
    currentDict: Record<string, FormattedApigwCustomDomain> = {},
    newDict: Record<string, FormattedApigwCustomDomain> = {},
  ) {
    const res = await this.apigw.client.DescribeServiceSubDomains({
      ServiceId: serviceId,
    });

    if ((res.Result?.DomainSet?.length ?? 0) > 0) {
      const DomainSet = res.Result?.DomainSet ?? [];
      // 解绑所有创建的自定义域名
      for (let i = 0; i < DomainSet.length; i++) {
        const domainItem = DomainSet[i];
        const domain = domainItem.DomainName ?? '';
        // 当前绑定状态与新的绑定状态一致，不解绑
        if (currentDict[domain] && deepEqual(currentDict[domain], newDict[domain])) {
          console.log(
            `Domain ${domainItem.DomainName} for service ${serviceId} unchanged, won't unbind`,
          );
          continue;
        }

        for (let j = 0; j < oldCustomDomains.length; j++) {
          // 只解绑由组件创建的域名
          if (oldCustomDomains[j].subDomain === domainItem.DomainName) {
            console.log(`Start unbind domain ${domainItem.DomainName} for service ${serviceId}`);
            await this.apigw.client.UnBindSubDomain({
              ServiceId: serviceId,
              SubDomain: domainItem.DomainName,
            });
          }
        }
      }
    }
  }

  /**
   * 为 API 网关服务绑定自定义域名
   *
   * @param root0 输入参数
   * @param root0.serviceId API 网关服务 ID
   * @param root0.subDomain API 网关子域名
   * @param root0.inputs API 网关绑定自定义域名输入
   * @returns 绑定成功后的输出
   */
  async bindCustomDomain({
    serviceId,
    subDomain,
    inputs,
  }: {
    serviceId: string;
    subDomain: string;
    inputs: ApigwBindCustomDomainInputs;
  }): Promise<ApigwBindCustomDomainOutputs[]> {
    console.log('Binding custom domain...');
    let { customDomains } = inputs;
    const { oldState = {} } = inputs;
    if (!customDomains) {
      // FIXME: 不存在自定义域名的时候，应该解绑之前绑定的域名
      customDomains = [];
    }

    const currentDict = await this.getCurrentCustomDomainsDict(serviceId);
    const newDict = getCustomDomainFormattedDict(inputs.customDomains ?? []);

    // 1. 解绑旧的自定义域名
    await this.unbindCustomDomain(serviceId, oldState?.customDomains ?? [], currentDict, newDict);

    // 2. bind user config domain
    const customDomainOutput: ApigwBindCustomDomainOutputs[] = [];
    if (customDomains && customDomains.length > 0) {
      console.log(`Start bind custom domain for service ${serviceId}`);
      for (let i = 0; i < customDomains.length; i++) {
        const domainItem = customDomains[i];
        const domainProtocol = domainItem.protocols
          ? formatProtocolString(domainItem.protocols)
          : inputs.protocols;
        const domainInputs = {
          serviceId,
          subDomain: domainItem.domain,
          netSubDomain: subDomain,
          certificateId: domainItem.certificateId,
          // default isDefaultMapping is true
          isDefaultMapping: domainItem.isDefaultMapping === false ? false : true,
          // if isDefaultMapping is false, should append pathMappingSet config
          pathMappingSet: domainItem.pathMappingSet || [],
          netType: domainItem.netType ?? 'OUTER',
          protocol: formatProtocolString(domainProtocol),
          isForcedHttps: domainItem.isForcedHttps === true,
        };

        try {
          const { domain } = domainItem;
          // 当前状态与新的状态一致，不进行绑定
          if (currentDict[domain] && deepEqual(currentDict[domain], newDict[domain])) {
            console.log(`Custom domain for service ${serviceId} unchanged, wont create.`);
            console.log(`Please add CNAME record ${subDomain} for ${domainItem.domain}.`);
          } else {
            await this.apigw.client.BindSubDomain(
              pascalCaseProperty({
                ...domainInputs,
              }),
            );
            console.log(`Custom domain for service ${serviceId} created successfullly.`);
            console.log(`Please add CNAME record ${subDomain} for ${domainItem.domain}.`);
          }

          customDomainOutput.push({
            isBinded: true,
            created: true,
            subDomain: domainItem.domain,
            cname: subDomain,
            url: `${domainProtocol.indexOf('https') !== -1 ? 'https' : 'http'}://${
              domainItem.domain
            }`,
          });
        } catch (e) {
          // User hasn't add cname dns record
          if (e.code === 'FailedOperation.DomainResolveError') {
            customDomainOutput.push({
              isBinded: false,
              subDomain: domainItem.domain,
              cname: subDomain,
              message: `您的自定义域名还未生效，请给域名 ${domainItem.domain} 添加 CNAME 记录 ${subDomain}，等待解析生效后，再次运行 'sls deploy' 完成自定义域名的配置`,
            });
          } else {
            throw e;
          }
        }
      }
    }

    return customDomainOutput;
  }
}

==> apigw-1.1.0.tgz: package/src/custom-domain.ts <==
import { getLogger, printToConsole } from '@jsonlog/logger';
import { pascalCaseProperty } from '@tencent-serverless-framework/common';
import { Apigw } from '.';
import { formatProtocolString } from './formatter';
import {
  ApigwBindCustomDomainInputs,
  ApigwBindCustomDomainOutputs,
  ApigwCustomDomain,
} from './interface';

const logger = getLogger({
  name: 'APIGW - CustomDomain',
  middlewares: [printToConsole],
});

/** 格式化的 API 网关自定义域名信息 */
interface FormattedApigwCustomDomain {
  /** 域名 */
  domain: string;
  /** 协议 */
  protocols: string;
  /** 证书 id */
  certificateId?: string;
  /** 是否是默认映射 */
  isDefaultMapping: boolean;
  /** 映射字典 */
  pathMappingSetDict: Record<string, string>;
  /** 网络类型 */
  netType: string;
}

/**
 * 判断是否是值类型
 *
 * @param obj 要判断的对象
 * @returns 是否是值类型
 */
export function isPrimitive(obj: any): boolean {
  return obj !== Object(obj);
}

/**
 * 深度比较对象是否相等
 *
 * @param obj1 对象 1
 * @param obj2 对象 2
 * @returns 是否相等
 */
export function deepEqual(obj1: any, obj2: any): boolean {
  if (obj1 === obj2) {
    return true;
  }

  if (isPrimitive(obj1) || isPrimitive(obj2)) {
    return obj1 === obj2;
  }

  if (Object.keys(obj1).length !== Object.keys(obj2).length) {
    return false;
  }

  // compare objects with same number of keys
  for (const key in obj1) {
    if (!(key in obj2)) {
      return false;
    }
    if (!deepEqual(obj1[key], obj2[key])) {
      return false;
    }
  }

  return true;
}

/**
 * 获取自定义域名的字典
 *
 * @param domains 自定义域名数组
 * @returns 自定义域名字典
 */
function getCustomDomainFormattedDict(domains: ApigwCustomDomain[]) {
  const domainDict: Record<string, FormattedApigwCustomDomain> = {};
  domains.forEach((d) => {
    const pmDict: Record<string, string> = {};
    for (const pm of d.pathMappingSet ?? []) {
      pmDict[pm.path] = pm.environment;
    }
    domainDict[d.domain] = {
      domain: d.domain,
      certificateId: d.certificateId ?? '',
      protocols: formatProtocolString(d.protocols ?? ''),
      isDefaultMapping: d.isDefaultMapping === false ? false : true,
      pathMappingSetDict: pmDict,
      netType: d.netType ?? 'OUTER',
    };
  });

  return domainDict;
}

/** API 网关组件自定义域名子模块 */
export class CustomDomain {
  /** API 网关组件 */
  apigw: Apigw;
  /**
   * 构造函数
   *
   * @param apigw API 网关组件（依赖注入）
   */
  constructor(apigw: Apigw) {
    this.apigw = apigw;
  }

  /**
   * 获取当前 API 自定义域名字典
   *
   * @param serviceId API ID
   * @returns 自定义域名字典
   */
  async getCurrentCustomDomainsDict(serviceId: string) {
    const res = await this.apigw.client.DescribeServiceSubDomains({
      ServiceId: serviceId,
    });

    const DomainSet = res.Result?.DomainSet;

    const domainDict: Record<string, FormattedApigwCustomDomain> = {};

    for (const d of DomainSet ?? []) {
      const domain: FormattedApigwCustomDomain = {
        domain: d.DomainName,
        protocols: d.Protocol,
        certificateId: d.CertificateId,
        isDefaultMapping: d.IsDefaultMapping,
        netType: d.NetType,
        pathMappingSetDict: {},
      };

      const res_ = await this.apigw.client.DescribeServiceSubDomainMappings({
        ServiceId: serviceId,
        SubDomain: d.DomainName,
      });

      const PathMappingSet = res_.Result?.PathMappingSet;

      PathMappingSet?.map((v) => {
        domain.pathMappingSetDict[v.Path] = v.Environment;
      });

      domainDict[domain.domain] = domain;
    }

    return domainDict;
  }

  /**
   * 解绑 API 网关所有由组件创建的自定义域名（通过查询 oldState）
   *
   * @param serviceId API 网关 ID
   * @param oldCustomDomains 状态中的自定义
   * @param currentDict 当前自定义域名字典
   * @param newDict 要绑定的自定义域名字典
   */
  async unbindCustomDomain(
    serviceId: string,
    oldCustomDomains: ApigwCustomDomain[],
    currentDict: Record<string, FormattedApigwCustomDomain> = {},
    newDict: Record<string, FormattedApigwCustomDomain> = {},
  ) {
    const res = await this.apigw.client.DescribeServiceSubDomains({
      ServiceId: serviceId,
    });

    if ((res.Result?.DomainSet?.length ?? 0) > 0) {
      const DomainSet = res.Result?.DomainSet ?? [];
      // 解绑所有创建的自定义域名
      for (let i = 0; i < DomainSet.length; i++) {
        const domainItem = DomainSet[i];
        const domain = domainItem.DomainName ?? '';
        // 当前绑定状态与新的绑定状态一致，不解绑
        if (currentDict[domain] && deepEqual(currentDict[domain], newDict[domain])) {
          logger.info({
            msg: `Domain ${domainItem.DomainName} for service ${serviceId} unchanged, won't unbind`,
          });
          continue;
        }

        for (let j = 0; j < oldCustomDomains.length; j++) {
          // 只解绑由组件创建的域名
          if (oldCustomDomains[j].subDomain === domainItem.DomainName) {
            logger.info({
              msg: `开始解绑自定义域名 ${domainItem.DomainName}，所属 API 网关 ${serviceId}`,
            });
            await this.apigw.client.UnBindSubDomain({
              ServiceId: serviceId,
              SubDomain: domainItem.DomainName,
            });
          }
        }
      }
    }
  }

  /**
   * 为 API 网关服务绑定自定义域名
   *
   * @param root0 输入参数
   * @param root0.serviceId API 网关服务 ID
   * @param root0.subDomain API 网关子域名
   * @param root0.inputs API 网关绑定自定义域名输入
   * @returns 绑定成功后的输出
   */
  async bindCustomDomain({
    serviceId,
    subDomain,
    inputs,
  }: {
    /** API 网关服务 ID */
    serviceId: string;
    /** API 网关子域名 */
    subDomain: string;
    /** API 网关绑定自定义域名输入 */
    inputs: ApigwBindCustomDomainInputs;
  }): Promise<ApigwBindCustomDomainOutputs[]> {
    logger.info({ msg: '正在绑定自定义域名 ...' });
    let { customDomains } = inputs;
    const { oldState = {} } = inputs;
    if (!customDomains) {
      // FIXME: 不存在自定义域名的时候，应该解绑之前绑定的域名
      customDomains = [];
    }

    const currentDict = await this.getCurrentCustomDomainsDict(serviceId);
    const newDict = getCustomDomainFormattedDict(inputs.customDomains ?? []);

    // 1. 解绑旧的自定义域名
    await this.unbindCustomDomain(serviceId, oldState?.customDomains ?? [], currentDict, newDict);

    // 2. bind user config domain
    const customDomainOutput: ApigwBindCustomDomainOutputs[] = [];
    if (customDomains && customDomains.length > 0) {
      logger.info({ msg: `开始绑定服务的自定义域名，服务 ID： ${serviceId}` });
      for (let i = 0; i < customDomains.length; i++) {
        const domainItem = customDomains[i];
        const domainProtocol = domainItem.protocols
          ? formatProtocolString(domainItem.protocols)
          : inputs.protocols;
        const domainInputs = {
          serviceId,
          subDomain: domainItem.domain,
          netSubDomain: subDomain,
          certificateId: domainItem.certificateId,
          // default isDefaultMapping is true
          isDefaultMapping: domainItem.isDefaultMapping === false ? false : true,
          // if isDefaultMapping is false, should append pathMappingSet config
          pathMappingSet: domainItem.pathMappingSet || [],
          netType: domainItem.netType ?? 'OUTER',
          protocol: formatProtocolString(domainProtocol),
          isForcedHttps: domainItem.isForcedHttps === true,
        };

        try {
          const { domain } = domainItem;
          // 当前状态与新的状态一致，不进行绑定
          if (currentDict[domain] && deepEqual(currentDict[domain], newDict[domain])) {
            logger.info({ msg: `服务 ${serviceId} 的自定义域名 ${domain} 没有变化，停止创建` });
            logger.info({ msg: `请在 ${domain} 的 CNAME 中填入  ${subDomain}` });
          } else {
            await this.apigw.client.BindSubDomain(
              pascalCaseProperty({
                ...domainInputs,
              }),
            );
            logger.info({ msg: `为服务 ${serviceId} 绑定自定义域名 ${domain} 成功` });
            logger.info({ msg: `请在 ${domain} 的 CNAME 中填入  ${subDomain}` });
          }

          customDomainOutput.push({
            isBinded: true,
            created: true,
            subDomain: domainItem.domain,
            cname: subDomain,
            url: `${domainProtocol.indexOf('https') !== -1 ? 'https' : 'http'}://${
              domainItem.domain
            }`,
          });
        } catch (e) {
          // User hasn't add cname dns record
          if (e.code === 'FailedOperation.DomainResolveError') {
            customDomainOutput.push({
              isBinded: false,
              subDomain: domainItem.domain,
              cname: subDomain,
              message: `您的自定义域名还未生效，请给域名 ${domainItem.domain} 添加 CNAME 记录 ${subDomain}，等待解析生效后，再次运行 'sls deploy' 完成自定义域名的配置`,
            });
          } else {
            throw e;
          }
        }
      }
    }

    return customDomainOutput;
  }
}
